@@ -1,26 +1,27 @@
-from typing import (
-    Any,
-    Optional,
-    List,
-    Iterable,
-    Union,
-    Dict,
-    Tuple,
-    Callable,
-    cast,
-    Sequence,
-)
-from warnings import warn
-import xarray as xr
+"""Native load and masking."""
 
-from odc.index import group_by_nothing, solar_offset
+from typing import (
+    Callable,
+    Dict,
+    Iterable,
+    List,
+    Optional,
+    Sequence,
+    Tuple,
+    Union,
+    cast,
+)
+
+import json
+import xarray as xr
+from pyproj import aoi, transformer
 from datacube import Datacube
 from datacube.model import Dataset
-from datacube.utils.geometry import GeoBox, gbox
-from datacube.api.core import output_geobox
 from datacube.testutils.io import native_geobox
-from odc.stac import dc_load
-from ._masking import _max_fuser, _nodata_fuser, enum_to_bool
+from datacube.utils.geometry import GeoBox, gbox
+
+from ._grouper import group_by_nothing, solar_offset
+from ._masking import _max_fuser, _nodata_fuser, _or_fuser, enum_to_bool, mask_cleanup
 from ._warp import xr_reproject
 
 
@@ -28,6 +29,8 @@ def compute_native_load_geobox(
     dst_geobox: GeoBox, ds: Dataset, band: str, buffer: Optional[float] = None
 ) -> GeoBox:
     """
+    Compute area of interest for a given Dataset given query.
+
     Take native projection and resolution from ``ds, band`` pair and compute
     region in that projection that fully encloses footprint of the
     ``dst_geobox`` with some padding. Construct GeoBox that encloses that
@@ -35,14 +38,16 @@ def compute_native_load_geobox(
 
     :param dst_geobox:
     :param ds: Sample dataset (only resolution and projection is used, not footprint)
-    :param band: Reference band to use (resolution of output GeoBox will match resolution of this band)
-    :param buffer: Buffer in units of CRS of ``ds`` (meters usually), default is 10 pixels worth
+    :param band: Reference band to use
+                 (resolution of output GeoBox will match resolution of this band)
+    :param buffer: Buffer in units of CRS of ``ds`` (meters usually),
+                   default is 10 pixels worth
     """
-
-    native = native_geobox(ds, basis=band)
+    native: GeoBox = native_geobox(ds, basis=band)
     if buffer is None:
-        buffer = 10 * cast(float, max(map(abs, native.resolution)))
+        buffer = 10 * cast(float, max(map(abs, native.resolution)))  # type: ignore
 
+    assert native.crs is not None
     return GeoBox.from_geopolygon(
         dst_geobox.extent.to_crs(native.crs).buffer(buffer),
         crs=native.crs,
@@ -51,6 +56,40 @@ def compute_native_load_geobox(
     )
 
 
+def choose_transform_path(
+    src_crs: str,
+    dst_crs: str,
+    transform_code: Optional[str] = None,
+    area_of_interest: Optional[Sequence[float]] = None,
+) -> str:
+    # leave gdal to choose the best option if nothing is specified
+    if transform_code is None and area_of_interest is None:
+        return {}
+
+    if area_of_interest is not None:
+        assert len(area_of_interest) == 4
+        area_of_interest = aoi.AreaOfInterest(*area_of_interest)
+
+    transformer_group = transformer.TransformerGroup(
+        src_crs, dst_crs, area_of_interest=area_of_interest
+    )
+    if transform_code is None:
+        return {"COORDINATE_OPERATION": transformer_group.transformers[0].to_proj4()}
+    else:
+        for t in transformer_group.transformers:
+            for step in json.loads(t.to_json()).get("steps", []):
+                if step.get("type", "") == "Transformation":
+                    authority_code = step.get("id", {})
+                    if transform_code.split(":")[0].upper() in authority_code.get(
+                        "authority", ""
+                    ) and transform_code.split(":")[1] == str(
+                        authority_code.get("code", "")
+                    ):
+                        return {"COORDINATE_OPERATION": t.to_proj4()}
+    # raise error if nothing is available
+    raise ValueError(f"Not able to find transform path by {transform_code}")
+
+
 def _split_by_grid(xx: xr.DataArray) -> List[xr.DataArray]:
     def extract(ii):
         yy = xx[ii]
@@ -62,6 +101,7 @@ def _split_by_grid(xx: xr.DataArray) -> List[xr.DataArray]:
     return [extract(ii) for ii in xx.groupby(xx.grid).groups.values()]
 
 
+# pylint: disable=too-many-arguments, too-many-locals
 def _load_with_native_transform_1(
     sources: xr.DataArray,
     bands: Tuple[str, ...],
@@ -74,6 +114,7 @@ def _load_with_native_transform_1(
     chunks: Optional[Dict[str, int]] = None,
     load_chunks: Optional[Dict[str, int]] = None,
     pad: Optional[int] = None,
+    **kwargs,
 ) -> xr.Dataset:
     if basis is None:
         basis = bands[0]
@@ -92,14 +133,16 @@ def _load_with_native_transform_1(
 
     if groupby is not None:
         if fuser is None:
-            fuser = _nodata_fuser
+            fuser = _nodata_fuser  # type: ignore
         xx = xx.groupby(groupby).map(fuser)
 
     _chunks = None
     if chunks is not None:
         _chunks = tuple(chunks.get(ax, -1) for ax in ("y", "x"))
 
-    return xr_reproject(xx, geobox, chunks=_chunks, resampling=resampling)
+    return xr_reproject(
+        xx, geobox, chunks=_chunks, resampling=resampling, **kwargs
+    )  # type: ignore
 
 
 def load_with_native_transform(
@@ -122,13 +165,15 @@ def load_with_native_transform(
     :param dss: A list of datasets to load
     :param bands: Which measurements to load
     :param geobox: GeoBox of the final output
-    :param native_transform: ``xr.Dataset -> xr.Dataset`` transform, should support Dask inputs/outputs
+    :param native_transform: ``xr.Dataset -> xr.Dataset`` transform,
+                             should support Dask inputs/outputs
     :param basis: Name of the band to use as a reference for what is "native projection"
     :param groupby: One of 'solar_day'|'time'|'idx'|None
     :param fuser: Optional ``xr.Dataset -> xr.Dataset`` transform
     :param resampling: Any resampling mode supported by GDAL as a string:
                        nearest, bilinear, average, mode, cubic, etc...
-    :param chunks: If set use Dask, must be in dictionary form ``{'x': 4000, 'y': 4000}``
+    :param chunks: If set use Dask, must be in dictionary form
+                   ``{'x': 4000, 'y': 4000}``
 
     :param load_chunks: Defaults to ``chunks`` but can be different if supplied
                         (different chunking for native read vs reproject)
@@ -139,6 +184,9 @@ def load_with_native_transform(
                 filter or morphological operators on input data.
 
     :param kw: Used to support old names ``dask_chunks`` and ``group_by``
+               also kwargs for reproject ``tranform_code`` in the form of
+               "authority:code", e.g., "epsg:9688", and ``area_of_interest``,
+               e.g., [-180, -90, 180, 90]
 
     1. Partition datasets by native Projection
     2. For every group do
@@ -154,31 +202,43 @@ def load_with_native_transform(
 
     if groupby is None:
         groupby = kw.get("group_by", "idx")
+
     if chunks is None:
         chunks = kw.get("dask_chunks", None)
 
-    sources = group_by_nothing(dss, solar_offset(geobox.extent))
-    _xx = [
-        _load_with_native_transform_1(
-            srcs,
-            bands,
-            geobox,
-            native_transform,
-            basis=basis,
-            resampling=resampling,
-            groupby=groupby,
-            fuser=fuser,
-            chunks=chunks,
-            load_chunks=load_chunks,
-            pad=pad,
+    sources = group_by_nothing(list(dss), solar_offset(geobox.extent))
+    _xx = []
+    # fail if the intended transform not available
+    # to avoid any unexpected results
+    for srcs in _split_by_grid(sources):
+        extra_args = choose_transform_path(
+            srcs.crs,
+            geobox.crs,
+            kw.get("transform_code"),
+            kw.get("area_of_interest"),
         )
-        for srcs in _split_by_grid(sources)
-    ]
+
+        _xx += [
+            _load_with_native_transform_1(
+                srcs,
+                tuple(bands),
+                geobox,
+                native_transform,
+                basis=basis,
+                resampling=resampling,
+                groupby=groupby,
+                fuser=fuser,
+                chunks=chunks,
+                load_chunks=load_chunks,
+                pad=pad,
+                **extra_args,
+            )
+        ]
 
     if len(_xx) == 1:
         xx = _xx[0]
     else:
-        xx = xr.concat(_xx, sources.dims[0])
+        xx = xr.concat(_xx, sources.dims[0])  # type: ignore
         if groupby != "idx":
             xx = xx.groupby(groupby).map(fuser)
 
@@ -203,7 +263,8 @@ def load_enum_mask(
 
     1. Load each mask time slice separately in native projection of the file
     2. Convert enum to Boolean (F:0, T:255)
-    3. Optionally (groupby='solar_day') group observations on the same day using OR for pixel fusing: T,F->T
+    3. Optionally (groupby='solar_day') group observations on the same day
+       using OR for pixel fusing: T,F->T
     4. Reproject to destination GeoBox (any resampling mode is ok)
     5. Optionally group observations on the same day using OR for pixel fusing T,F->T
     6. Finally convert to real Bool
@@ -262,36 +323,39 @@ def load_enum_filtered(
     :param geobox: GeoBox of the final output
     :param categories: Enum values or names
 
-    :param filters: iterable tuples of morphological operations in the order you want them to perform
-                    e.g. [("opening", 2), ("dilation", 5)]
+    :param filters: iterable tuples of morphological operations in the order
+                    you want them to perform, e.g., [("opening", 2), ("dilation", 5)]
     :param groupby: One of 'solar_day'|'time'|'idx'|None
     :param resampling: Any resampling mode supported by GDAL as a string:
                        nearest, bilinear, average, mode, cubic, etc...
-    :param chunks: If set use Dask, must be in dictionary form ``{'x': 4000, 'y': 4000}``
+    :param chunks: If set use Dask, must be in dictionary form
+                   ``{'x': 4000, 'y': 4000}``
     :param kw: Passed on to ``load_with_native_transform``
 
 
     1. Load each mask time slice separately in native projection of the file
     2. Convert enum to Boolean
-    3. Optionally (groupby='solar_day') group observations on the same day using OR for pixel fusing: T,F->T
+    3. Optionally (groupby='solar_day') group observations on the same day
+       using OR for pixel fusing: T,F->T
     4. Optionally apply ``mask_cleanup`` in native projection (after fusing)
     4. Reproject to destination GeoBox (any resampling mode is ok)
     5. Optionally group observations on the same day using OR for pixel fusing T,F->T
     """
-    from ._masking import mask_cleanup, _or_fuser
 
     def native_op(xx: xr.Dataset) -> xr.Dataset:
         _xx = enum_to_bool(xx[band], categories)
         return xr.Dataset(
-            {band: _xx}, attrs={"native": True},  # <- native flag needed for fuser
+            {band: _xx},
+            attrs={"native": True},  # <- native flag needed for fuser
         )
 
     def fuser(xx: xr.Dataset) -> xr.Dataset:
         """
+        Fuse with OR.
+
         Fuse with OR, and when fusing in native pixel domain apply mask_cleanup if
         requested
         """
-
         is_native = xx.attrs.get("native", False)
         xx = xx.map(_or_fuser)
         xx.attrs.pop("native", None)
@@ -304,10 +368,10 @@ def load_enum_filtered(
         return xx
 
     # unless set by user to some value use largest filter radius for pad value
-    pad = kw.pop("pad", None)
+    pad: Optional[int] = kw.pop("pad", None)
     if pad is None:
         if filters is not None:
-            pad = max(list(zip(*filters))[1])
+            pad = max(list(zip(*filters))[1])  # type: ignore
 
     xx = load_with_native_transform(
         dss,
